/**
    Copyright (C) 1997 Pekka Nikander & Juha Prssinen.
    All rights Reserved.

    XXX   %Z%%M% version %I% date:%G%

    Module:  JacobDatagramSocket (implementation)
    Package: inet
    File:    JacobDatagramSocket.java


*/
package fi.hut.tcm.jacob.tcpip;

import fi.hut.tcm.jacob.tcpip.udp.*;
import fi.hut.tcm.jacob.tcpip.icmp.*;
import fi.hut.tcm.jacob.tcpip.ip.*;
import fi.hut.tcm.jacob.tcpip.arp.*;
import fi.hut.tcm.jacob.tcpip.ethernet.*;

import fi.hut.tcm.jacob.conduits.*;
import java.util.Enumeration;
import java.util.Vector;
import java.net.*;
import java.io.*;

/**
    JacobDatagramSocket
    
    
*/
public final class JacobDatagramSocket
  extends Object {
    
  private final SimpleScheduler scheduler = SimpleScheduler.instance;
  
  public static final int UDP_PROTOCOL_ID = 17; // Protocol field in IP header
  private long localEthAddr;
  private byte[] localIpAddr = new byte[4];

  private UdpAdaptor adaptor;
  private DatagramPacket packet;
  private boolean packetReceived;

  private static final String itblCfgFilename = new String("itbl.cfg");
  private static final String rtblCfgFilename = new String("rtbl.cfg");

  public JacobDatagramSocket(int port, InetAddress addr) 
  throws java.net.SocketException {

    localIpAddr = addr.getAddress();
    int localPort = port;

    /*************
     * Instances *
     *************/

    // Init interface table and routing table

    TcpipInterfaceTable ifTable = null;
    TcpipRoutingTable rTable = null;
    try {
      ifTable = new TcpipInterfaceTable(itblCfgFilename);
      rTable = new TcpipRoutingTable(rtblCfgFilename);
    } catch(IOException e) {
      throw new SocketException("I/O Error reading config files");
    } catch(IllegalArgumentException e) {
      throw new SocketException("Syntax error in config files");
    }

    String ifId = null;
    TcpipInterfaceEntry entry = null;
    Enumeration interfaces = ifTable.listEntries();

    int localIpInt =
      ((int) localIpAddr[0] & 0xff) << 24 |      
      ((int) localIpAddr[1] & 0xff) << 16 |      
      ((int) localIpAddr[2] & 0xff) <<  8 |      
      ((int) localIpAddr[3] & 0xff) <<  0;

    while (interfaces.hasMoreElements()) {
      entry = (TcpipInterfaceEntry) interfaces.nextElement();
      if (entry.getDefIpAddr() == localIpInt) {
	ifId = entry.getId();
	break;
      }
    }
    if (ifId == null) {
      System.err.println("Unable to get interface");
      System.exit(-1);
    }
    
    localEthAddr = entry.getEthAddr();
    int mtu = entry.getMtu();

    // Ethernet

    /*
    // pseudo, machine-internal ethernet:
    PseudoEthernetAdaptor ethAdapt = new PseudoEthernetAdaptor(EthernetAddress.toArray(localEthAddr));
      */

    // Real ethernet:
    Vector filteredIps = new Vector();
    filteredIps.addElement(new IpAddress("130.233.45.26")); // clotam
    filteredIps.addElement(new IpAddress("130.233.45.255")); // broadcast
    filteredIps.addElement(new IpAddress("130.233.195.255")); // broadcast
    EthernetAdaptor ethAdapt =
      new EthernetAdaptor( new EthernetAddress(localEthAddr),
			   filteredIps,
			   "/dev/le", 0 );

    // common to real and pseudo-ethernet:

    EthernetHeaderSession ethHdrSes = 
      new EthernetHeaderSession(EthernetHeaderState.instance, localEthAddr);

    // ARP

    ArpTable arpTable = new ArpTable();

    EtherAddrAccessor etherAddrAccessor = new EtherAddrAccessor(arpTable);
    Mux etherAddrMux = new Mux(etherAddrAccessor);
    ArpIcmpAdaptor arpIcmpAdapt = new ArpIcmpAdaptor();
    ArpReplyAdaptor arpReplyAdapt = 
      new ArpReplyAdaptor(arpTable, ifTable, ifId);
    ArpRequestSession arpRequestSes = 
      new ArpRequestSession(arpTable, ArpRequestState.instance, ifTable);
    ArpSelectorAccessor arpSelectorAccessor = new ArpSelectorAccessor();
    Mux arpSelectorMux = new Mux(arpSelectorAccessor);
    Terminator arpSelectorMuxTer = new Terminator(); 

    // IP

    IpChecksumSession ipCSes =
      new IpChecksumSession(IpChecksumState.instance);
    IpFragmentationSession ipFrSes =
      new IpFragmentationSession(IpFragmentationState.instance, ifTable);
    IpRoutingAccessor ipRAcc = new IpRoutingAccessor();
    Mux ipRMux = new Mux(ipRAcc);
    Terminator ipRTer = new Terminator();
    IpRoutingSession ipRSes =
      new IpRoutingSession(IpRoutingOffState.instance, rTable, ifTable);
    IpReassemblySession ipReSes = 
      new IpReassemblySession(IpReassemblyState.instance);
    IpProtocolAccessor ipPAcc = new IpProtocolAccessor();
    Mux ipPMux = new Mux(ipPAcc);
    Terminator ipPTer = new Terminator();
    IpHeaderSession ipHSes = new IpHeaderSession(IpHeaderState.instance);

    // IP/UDP test
    // IpTestAdaptor ipTAdap = new IpTestAdaptor(localEthAddr, mtu);

    // UDP

    UdpChecksumSession udpCSes =
      new UdpChecksumSession(UdpChecksumState.instance, rTable, ifTable);
    UdpPortAccessor udpPAcc = new UdpPortAccessor();
    Mux udpPMux = new Mux(udpPAcc);
    Terminator udpPTer = new Terminator(); 
    UdpHeaderSession udpHSes = 
      new UdpHeaderSession(UdpHeaderState.instance, localPort);
    
    // Application

    adaptor = new UdpAdaptor(this);
    
    // Stub sessions
    TcpipStubSession ipStubSes =
      new TcpipStubSession(null,
			   IpChecksumFromNetMessenger.instance,
			   /* IpStubMessenger.instance */
			   null
			   );
    TcpipStubSession udpStubSes =
      new TcpipStubSession(IpHeaderToNetMessenger.instance,
			   UdpChecksumFromNetMessenger.instance, 
			   /* UdpStubMessenger.instance */
			   null
			   );

    // Protocol conduits
    IP ipProt = new IP(ipCSes, ipHSes);
    UDP udpProt = new UDP(udpCSes, udpHSes);

    /*********************
     * Conduit structure *
     *********************/

    try {
      
      ethAdapt.attach(Side.A, ethHdrSes);
      ethHdrSes.attach(Side.A, ethAdapt);

      ethHdrSes.attach(Side.B, arpSelectorMux);

      // ARP

      arpSelectorMux.attach(Side.A, ethHdrSes);
      arpSelectorMux.attach("1", etherAddrMux);
      arpSelectorMux.attach("2", arpReplyAdapt);
      arpSelectorMux.attach("3", arpRequestSes);
      arpSelectorMux.attach(Side.B, arpSelectorMuxTer);

      arpReplyAdapt.attach(Side.A, arpSelectorMux);
      arpSelectorMuxTer.attach(Side.A, arpSelectorMux);
            
      etherAddrMux.attach(Side.A, ipStubSes);
      etherAddrMux.attach("1", arpSelectorMux);
      etherAddrMux.attach("2", arpRequestSes);
      etherAddrMux.attach(Side.B, arpIcmpAdapt);

      arpRequestSes.attach(Side.A, arpSelectorMux);
      arpRequestSes.attach(Side.B, etherAddrMux);

      arpIcmpAdapt.attach(Side.A, etherAddrMux);


      // IP/UDP test 
      //ipTAdap.attach(Side.A, ipStubSes);
      //ipStubSes.attach(Side.A, ipTAdap);

      // ARP-IP Stub

      ipStubSes.attach(Side.A, etherAddrMux);
      ipStubSes.attach(Side.B, ipProt);

      // IP

      ipProt.attach(Side.A, ipStubSes);

      ipCSes.attach(Side.B, ipFrSes);

      ipFrSes.attach(Side.A, ipCSes);
      ipFrSes.attach(Side.B, ipRMux);

      ipRMux.attach(Side.B, ipRTer);
      ipRMux.attach(ifId, ipFrSes);
      // ipRMux.attach("eth0", ipFrSes);
      // ipRMux.attach("eth1", ipFrSes);
      ipRMux.attach(Side.A, ipRSes);
      
      ipRSes.attach(Side.A, ipRMux);
      ipRSes.attach(Side.B, ipReSes);

      ipReSes.attach(Side.A, ipRSes);
      ipReSes.attach(Side.B, ipPMux);

      ipPMux.attach(Side.A, ipReSes);
      ipPMux.attach(new Byte((byte) UDP_PROTOCOL_ID), ipHSes);
      ipPMux.attach(Side.B, ipPTer);

      ipPTer.attach(Side.A, ipPMux);

      ipHSes.attach(Side.A, ipPMux);

      ipProt.attach(Side.B, udpStubSes);

      // IP-UDP Stub

      udpStubSes.attach(Side.A, ipProt);
      udpStubSes.attach(Side.B, udpProt);

      // UDP

      udpProt.attach(Side.A, udpStubSes);

      udpCSes.attach(Side.B, udpPMux);

      udpPMux.attach(Side.A, udpCSes);
      udpPMux.attach(new Integer(localPort), udpHSes);
      udpPMux.attach(Side.B, udpPTer);

      udpHSes.attach(Side.A, udpPMux);

      udpProt.attach(Side.B, adaptor);

      // Application

      adaptor.attach(Side.A, udpProt);

      // IP/UDP test
      // ipTAdap.start();

    } catch (SideException e) {
      System.err.println("Host " + e + " failed to initialize");
      System.exit(2);
    }
  }

  public void send(DatagramPacket p)
  throws java.io.IOException {
    adaptor.send(p);
  }

  public synchronized void receive(DatagramPacket p) 
  throws java.io.IOException {
    while (!packetReceived) {
      try {
	wait();
      } catch (InterruptedException e) {}
    }
    p.setAddress(packet.getAddress());
    p.setPort(packet.getPort());
    p.setData(packet.getData());
    p.setLength(packet.getLength());
    packetReceived = false;
  }

  public synchronized void put(DatagramPacket p) {
    packet = p;
    packetReceived = true;
    notifyAll();
  }

  public byte[] getLocalIpAddr() {
    byte[] newIpAddr = new byte[4];
    System.arraycopy(localIpAddr, 0, newIpAddr, 0, 4);
    return newIpAddr;    
  }

  public void close() {
  }

}; // JacobDatagramSocket



