/**
    Copyright (C) 1997 Pekka Nikander & Juha Prssinen.
    All rights Reserved.

    XXX   %Z%%M% version %I% date:%G%

    Module:  TcpipInterfaceTable (implementation)
    Package: tcpip
    File:    TcpipInterfaceTable.java


*/
package fi.hut.tcm.jacob.tcpip;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;
import java.io.*;

/**
    TcpipInterfaceTable
    
    The InterfaceTable contains the following information for each existing
    network interface:
    Interface name (Interface1, Interface2, ...)
    IP address(es) and netmask
    MTU
    
*/
public class TcpipInterfaceTable {
    private Hashtable entries;
    protected static final int PARSE_ID = 0;
    protected static final int PARSE_ETHDIGIT = 1;
    protected static final int PARSE_ETHDOT = 2;
    protected static final int PARSE_MTU = 3;
    protected static final int PARSE_IPDIGIT = 4;
    protected static final int PARSE_IPDOT = 5;

    public TcpipInterfaceTable() {
    //+ TcpipInterfaceTable::TcpipInterfaceTable( )
      entries = new Hashtable();
    //- TcpipInterfaceTable::TcpipInterfaceTable( )
    }

    public TcpipInterfaceTable(String configFilename)
    throws IOException, IllegalArgumentException {
    //+ TcpipInterfaceTable::TcpipInterfaceTable(String)

      entries = new Hashtable();


      StreamTokenizer tokenizer;
      FileInputStream cfgFile;
      try {
	cfgFile = new FileInputStream(configFilename);
	tokenizer = new StreamTokenizer(new InputStreamReader(cfgFile));
      } catch (FileNotFoundException e) {
	throw new FileNotFoundException(configFilename);
      }

      // Init tokenizer
      tokenizer.resetSyntax();
      tokenizer.wordChars('0', '9');
      tokenizer.wordChars('a', 'z');
      tokenizer.wordChars('A', 'Z');
      tokenizer.commentChar('#');
      tokenizer.whitespaceChars(0, ' ');
      tokenizer.eolIsSignificant(true);

      int tokmode = PARSE_ID;
      boolean satisfied = false;
      String newId = null;
      long newEthAddr = 0;
      int ethOffs = 0;
      int newMtu = 0;
      Vector newIpAddrs = null;
      int newIpAddr = 0;
      int ipOffs = 0;
      int value = 0;

      // Parse configuration file and create entries
      try {
	while (tokenizer.nextToken() != StreamTokenizer.TT_EOF) {

	  switch (tokmode) {
	    
	  case PARSE_ETHDIGIT:
	  case PARSE_MTU:
	  case PARSE_IPDIGIT:
	    if (tokenizer.ttype == StreamTokenizer.TT_WORD) {
	      try {
		value = new Integer(tokenizer.sval).intValue();
	      } catch (NumberFormatException e) {
		throw new IllegalArgumentException();
	      }
	    }
	  }
	    
	  switch (tokmode) {

	  case PARSE_ID:
	    if (tokenizer.ttype == StreamTokenizer.TT_WORD) {
	      newId = new String(tokenizer.sval);
	      tokmode = PARSE_ETHDIGIT;
	      satisfied = false;
	    } else if (tokenizer.ttype == StreamTokenizer.TT_EOL) {
	    } else {
	      throw new IllegalArgumentException();
	    }
	    break;

	  case PARSE_ETHDIGIT:
	    if (tokenizer.ttype == StreamTokenizer.TT_WORD) {
	      if (value > 0xff) {
		throw new IllegalArgumentException();
	      }
	      newEthAddr += (long) value << (40 - ethOffs);
	      if ((ethOffs += 8) < 48) {
		tokmode = PARSE_ETHDOT;
	      } else {
		ethOffs = 0;
		tokmode = PARSE_MTU;
	      }
	    } else {
	      throw new IllegalArgumentException();
	    }
	    break;
	    
	  case PARSE_ETHDOT:
	    if (tokenizer.ttype == '.') {
	      tokmode = PARSE_ETHDIGIT;
	    } else {
	      throw new IllegalArgumentException();
	    }
	    break;
	    
	  case PARSE_MTU:
	    if (tokenizer.ttype == StreamTokenizer.TT_WORD) {
	      newMtu = value;
	      tokmode = PARSE_IPDIGIT;
	    } else {
	      throw new IllegalArgumentException();
	    }
	    break;
	    
	  case PARSE_IPDIGIT:
	    if (tokenizer.ttype == StreamTokenizer.TT_WORD) {
	      if (value > 0xff) {
		throw new IllegalArgumentException();
	      }
	    
	      if (newIpAddrs == null) {
		newIpAddrs = new Vector();
	      }
	      newIpAddr += (int) value << (24 - ipOffs);
	      if ((ipOffs += 8) < 32) {
		tokmode = PARSE_IPDOT;
	      } else {
		ipOffs = 0;
		newIpAddrs.addElement(new Integer(newIpAddr));
		newIpAddr = 0;
	      }
	  
	    } else if (tokenizer.ttype == StreamTokenizer.TT_EOL) {
	      if (ipOffs > 0 || newIpAddrs == null) {
		throw new IllegalArgumentException();
	      }
	      // An entry has been parsed
	      addEntry(new TcpipInterfaceEntry(newId, 
					       newEthAddr, 
					       newMtu,
					       newIpAddrs));
	      newEthAddr = 0;
	      newIpAddr = 0;
	      newIpAddrs = null;
	      satisfied = true;
	      tokmode = PARSE_ID;
	    } else {
	      throw new IllegalArgumentException();
	    }
	    break;
	    
	  case PARSE_IPDOT:
	    if (tokenizer.ttype == '.') {
	      tokmode = PARSE_IPDIGIT;
	    } else {
	      throw new IllegalArgumentException(tokenizer.sval);
	    }
	    break;
	    
	  }
	}
      } catch (IOException e) {
	throw new IOException();
      }

      if (!satisfied) {
	throw new IOException();
      }

      cfgFile.close();

      printEntries();

    //- TcpipInterfaceTable::TcpipInterfaceTable(String)
    }

    public TcpipInterfaceEntry get(int ipAddr) {
    //+ TcpipInterfaceTable::get(int)
      Enumeration ents = entries.elements();
      while (ents.hasMoreElements()) {
	TcpipInterfaceEntry entry = (TcpipInterfaceEntry) ents.nextElement();
	if (entry.hasIpAddr(ipAddr)) {
	  return entry;
	}
      }
      return null;
    //- TcpipInterfaceTable::get(int)
    }

    public TcpipInterfaceEntry get(String id) {
    //+ TcpipInterfaceTable::get(String)
      return (TcpipInterfaceEntry) entries.get(id);
    //- TcpipInterfaceTable::get(String)
    }

    public Enumeration listEntries() {
    //+ TcpipInterfaceTable::listEntries( )
      return entries.elements();
    //- TcpipInterfaceTable::listEntries( )
    }

    public void printEntries() {
    //+ TcpipInterfaceTable::printEntries( )
      Enumeration ents = entries.elements();
      while (ents.hasMoreElements()) {
	System.err.println();
	((TcpipInterfaceEntry) ents.nextElement()).print();
	System.err.println();
      }
    //- TcpipInterfaceTable::printEntries( )
    }

    public TcpipInterfaceEntry addEntry(TcpipInterfaceEntry entry) {
    //+ TcpipInterfaceTable::addEntry(TcpipInterfaceEntry)
      return (TcpipInterfaceEntry) entries.put(entry.id, entry);
    //- TcpipInterfaceTable::addEntry(TcpipInterfaceEntry)
    }

    public TcpipInterfaceEntry removeEntry(String id) {
    //+ TcpipInterfaceTable::removeEntry(String)
      return (TcpipInterfaceEntry) entries.remove(id);
    //- TcpipInterfaceTable::removeEntry(String)
    }

}; //TcpipInterfaceTable
