/**
    Copyright (C) 1997 Pekka Nikander & Juha Prssinen.
    All rights Reserved.

    XXX   %Z%%M% version %I% date:%G%

    Module:  ArpReplyAdaptor (implementation)
    Package: arp
    File:    ArpReplyAdaptor.java


*/
package fi.hut.tcm.jacob.tcpip.arp;
import fi.hut.tcm.jacob.conduits.*;
import fi.hut.tcm.jacob.tcpip.*;

/**
    ArpReplyAdaptor
    
    - Receives ARP requests from the ARPSelectorMUX and generates ARP
    replies for those requests asking for
         our ethernet address. <br>
    - As an optimization, the ARP Reply Adaptor may update the ARPTable
    with the source address information in
         the ARP request. 
    
    
*/
public class ArpReplyAdaptor
  extends Adaptor {
    protected TcpipInterfaceTable ifTable;
    protected String ifKey;
    protected ArpTable arpTable;

    /**
        Makes the translation for messages from
        ARP request to ARP reply and sends it
        back to ArpSelectorMUX.
        
        @param message
    */
    public void adapt(Message message) {
    //+ ArpReplyAdaptor::adapt(Message)
      int SIZEOFARPMESSAGE = 28;
      long tmpl;
      int tmpi;

      TcpipOutOfBand oob = (TcpipOutOfBand) message.getOob();
      Object buf = message.getContents();
      int len = UtilMsgr.i.getLength(buf);

      long ethSrc;    // These come from the real ethernet header
      long ethDst;
      int frameType;

      long sender_ha;  // These are the fields of the ARP request,
      int  sender_ip;  // that is, in the ethernet data portion
      long target_ha;
      int  target_ip;

      ethSrc = oob.getSrcEthAddr();
      ethDst = oob.getDestEthAddr();
      frameType = oob.getEthFrameType();

      boolean goodRequest = true;
      boolean shouldAnswer = false;

      if ( len < SIZEOFARPMESSAGE )
	goodRequest = false;

      // 0x0001 = 10 Mbit Ethernet, 0x0800 = DOD IPv4
      if ( UtilMsgr.i.getIntUnaligned(buf, 0) != 0x00010800)
	goodRequest = false;

      // byte length of each hardware address = 6, protocol address = 4
      if ( UtilMsgr.i.getShortUnaligned(buf, 4) != 0x0604 )
	goodRequest = false;

      // ARP Request = 0x0001
      if ( UtilMsgr.i.getShortUnaligned(buf, 6) != 0x0001 )
	goodRequest = false;

      if ( ! goodRequest) {
	if (Arp.DebugLevel > 0) {  // #Debug#
	  System.out.println("Bad ARP request detected.");
	}
	return;
      }
      
      sender_ha = UtilMsgr.i.get6BytesUnaligned(buf, 8);
      sender_ip = UtilMsgr.i.getIntUnaligned(buf, 14);
      target_ha = UtilMsgr.i.get6BytesUnaligned(buf, 18);
      target_ip = UtilMsgr.i.getIntUnaligned(buf, 24);
      if (sender_ha != ethSrc) {
	goodRequest = false;   // Something very fishy going on here..
      }
      if (Arp.DebugLevel > 0) {  // #Debug#
	System.out.print("ARP: Who-has " + IpAddress.toString(target_ip) +
			 " Tell " + IpAddress.toString(sender_ip));
      }
      if (goodRequest) {
	//  Code below == retrieveMyEtherAddr( target_ip, target_ha );
	TcpipInterfaceEntry ifEntry = ifTable.get( target_ip );
	if ( ifEntry != null ) {
	  target_ha = ifEntry.getEthAddr();
	  shouldAnswer = true;
	}
      }
      
      if ( ! shouldAnswer ) {
	// The request was malformed, or not intended for us.
	if (Arp.DebugLevel > 0) { // #Debug#
	  System.out.println(", I won't answer.");
	}
	return;
      }
      
      if (Arp.DebugLevel > 0) {  // #Debug#
	System.out.println(", --> I'll answer this ARP request!");
      }
      
      // Now we have decided to answer the request
      // As an optimisation, remember the
      // address of the requestor, because we will probably be
      // talking to it very soon.
      arpTable.insertEntry(sender_ip, sender_ha, Arp.ARPOK);
      
      // Convert into an ARP reply:
      tmpi = sender_ip;
      sender_ip = target_ip;
      target_ip = tmpi;
      tmpl = sender_ha;
      sender_ha = target_ha;
      target_ha = tmpl;
      
      // Change to an ARP response. The MUX's accessor must either leave
      // this alone on the way out, or be able to do this change also:
      UtilMsgr.i.setShortUnaligned(buf, 6, 0x0002);
      // Put back into the packet, in the correct places:
      UtilMsgr.i.set6BytesUnaligned(buf, 8, sender_ha);
      UtilMsgr.i.setIntUnaligned(buf, 14, sender_ip);
      UtilMsgr.i.set6BytesUnaligned(buf, 18, target_ha);
      UtilMsgr.i.setIntUnaligned(buf, 24, target_ip);
      
      //      oob.setSrcEthAddr(ethDst);  // BUG! with this source will be Broadcast!
      oob.setSrcEthAddr(sender_ha);       // This should be ok...
      oob.setDestEthAddr(ethSrc);
      
      message.setMessenger(EthernetHeaderToNetMessenger.instance);
      MessageTransporter mt =
	MessageTransporter.newMessageTransporter(sides[Side.A], message);
      SimpleScheduler.instance.makeRunnable(mt);
      
      if (Arp.DebugLevel > 0) {  // #Debug#
	System.out.println("--> Sent an ARP reply:");
	System.out.print("    Dst: IP = " + IpAddress.toString(target_ip));
	System.out.println("  MAC = " + EthernetAddress.toString(target_ha));
	
	System.out.print("    Src: IP = " + IpAddress.toString(sender_ip));
	System.out.println("  MAC = " + EthernetAddress.toString(sender_ha));
      }
      
      if (Arp.DebugLevel > 1) {  // #Debug#
	System.out.println("Packet dump of ARP reply:");
	UtilMsgr.i.hexDump(buf);
      }
	
    //- ArpReplyAdaptor::adapt(Message)
    }

    /**
        IfKey is the key for this ethernet interface in the interface
        table.
    */
    public ArpReplyAdaptor(ArpTable arpTable, TcpipInterfaceTable tablePtr, String interfaceKey) {
    //+ ArpReplyAdaptor::ArpReplyAdaptor(ArpTable, TcpipInterfaceTable, String)
      super();          // Do this explicitly
      ifTable = tablePtr;
      ifKey = interfaceKey;
      this.arpTable = arpTable;
    //- ArpReplyAdaptor::ArpReplyAdaptor(ArpTable, TcpipInterfaceTable, String)
    }

}; //ArpReplyAdaptor

