/*
 * File: EthernetAdaptorImpl.c -- Conduits Ethernet adapter for Solaris
 *
 * Managing the use of /dev/le
 *
 * Authors: Bengt Sahlin <bos@tcm.hut.fi>
 *          Pekka Nikander <pnr@tcm.hut.fi>
 *
 * Created: Mon Jul 21 14:53:38 1997
 *
 * Corrections by Mikael Suokas <csuokas@cc.hut.fi>, 1998
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/stropts.h>
#include <sys/dlpi.h>
#include <sys/ethernet.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include "EthernetAdaptor.h"
#include <stdio.h>		/* Define BUFSIZ and NULL */

/*
 * The Ethernet device driver, e.g. /dev/le, can be used from user space to
 * receive and send ethernet packets (root permissions is needed, however).
 * The driver is accessed through DLPI interface, using DLPI primitives.
 * It can be opened like a normal driver with the open() system call.
 * To use the driver, it has to be attached and bound. In bound state
 * one can receive packets from and send packets to the Ethernet.
 *
 * The ppa (Physical Point of Attachment) indicates the corresponding
 * device instance unit number (usually 0 on first open). The sap 
 * (Service Access Point) indicates the type of service used. For example,
 * using sap value ETHERTYPE_IP (defined in /sys/ethernet.h) allows one
 * to send and receive IP packets. To associate a ppa, a DL_ATTACH_REQ
 * is issued, to bind a sap, a DL_BIND_REQ is issued.
 *
 * The packets are sent and received using the DLPI primitives
 * DL_UNITDATA_REQ and DL_UNITDATA_IND. 
 *
 * For more information on /dev/le, consult the le man page.
 * For more information on dlpi, consult the dlpi man page.
 * Refer to SUNS Streams manuals for more information on STREAMS
 */

/* Source address, destination address and ethernet types were
 * handled wrong. The fix is very much a trial-and-error one.
 * The constant below should be defined in some oter way, for instance.
 * Grep the source for XXX if you really wish to see the most suspect
 * code.
 */

#define DL_ETHER_TYPE_OFFSET 34


/*
 * Local type definitions
 */

typedef struct unitdata_packet
{
  dl_unitdata_req_t unitdata_req;
  struct ether_addr eth_addr;
  u_short ether_type;
} udata_packet_t;

/*
 * C level functions
 */

static int  ethernet_open(char *string, int ppa, int sap, int promisc);
static void ethernet_close(int fd);
static int  ethernet_close_internal(int fd, int state);
static int  ethernet_send(int fd, int ethertype, struct ether_addr *addr,
			  unsigned char *buffer, int length);
static int  ethernet_receive(int fd, unsigned short *ethertype,
			     struct ether_addr *addrs,
			     unsigned char *buffer, int maxlength);
static int  ethernet_address(int fd, struct ether_addr *addr);

/*
 * DLPI functions
 */

static int dlpi_attach_req(int fd, int ppa);
static int dlpi_bind_req(int fd, int sap);
static int dlpi_info_req(int fd);
static int dlpi_phys_addr_req(int fd);
static int dlpi_promiscon_req(int fd);

static int dlpi_detach_req(int fd);
static int dlpi_unbind_req(int fd);
static int dlpi_recv_ack(int fd, int okprim, union DL_primitives *rcvctl,
			 char *rcvdata, int *rcvdatalen);

/*
 * Debugging functions
 */

static void print_phys_addr(int fd);
static void print_le_info(int fd);
static void print_packet(unsigned char *ptr, int len, 
			 unsigned char *ptr2, int clen);


/* 
 * The Java interface
 */

long 
fi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor_open(
  struct Hfi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor *adaptor_handle,
  HArrayOfByte *device_handle,
  long sap,
  long ppa,
  /*boolean*/ long promisc) 
{
  ArrayOfByte *device = unhand(device_handle);
  return ethernet_open(device->body, sap, ppa, promisc);

}

void 
fi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor_close(
  struct Hfi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor *adaptor_handle,
  long handle)
{
  ethernet_close(handle);
}

long
fi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor_send(
  struct Hfi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor *adaptor_handle,
  HArrayOfByte *buffer_handle,
  long offset,
  long length)
{
  Classfi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor *adaptor = 
    unhand(adaptor_handle);
  int fd        = adaptor->handle;
  int ethertype = adaptor->sap;
  ArrayOfByte *buffer = unhand(buffer_handle);

  if (offset < 0 || length < 0)
    SignalError(0, "java/lang/ArrayIndexOutOfBoundsException",
		"offset < 0 || length < 0");

  if (obj_length(buffer_handle) < offset + length)
    SignalError(0, "java/lang/ArrayIndexOutOfBoundsException",
		"offset + length > buffer.length");

  /* OLD CODE
  return ethernet_send(fd, 
		       ethertype, 
		      (struct ether_addr*)(buffer->body + offset + ETHERADDRL),
		       buffer->body + offset + sizeof(struct ether_header), 
		       length);
  */

  /* The ethernet header is 6 bytes src, 6 bytes dst, 2 bytes type */
  ethertype = *((unsigned short *) (buffer->body + offset + 12)); 

  return ethernet_send(fd, 
		       ethertype,
		       (struct ether_addr*)(buffer->body + offset),
		       buffer->body + offset + sizeof(struct ether_header), 
		       length);
}

long 
fi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor_receive(
  struct Hfi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor *adaptor_handle,
  HArrayOfByte *buffer_handle,
  long offset,
  long maxlength) 
{
  Classfi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor *adaptor = 
    unhand(adaptor_handle);
  int fd = adaptor->handle;
  ArrayOfByte *buffer = unhand(buffer_handle);
  /* Test addition by M.S: */
  unsigned short ether_type;

  if (offset < 0 || maxlength < 0)
    SignalError(0, "java/lang/ArrayIndexOutOfBoundsException",
		"offset < 0 || maxlength < 0");

  /* test code: sizeof( struct ether_header) */
  if (obj_length(buffer_handle) < offset + maxlength + sizeof(struct ether_header) )
    SignalError(0, "java/lang/ArrayIndexOutOfBoundsException",
		"offset + maxlength + sizeof(struct ether_header) > buffer.length");

  return ethernet_receive(fd, 
			  (unsigned short *)(buffer->body + offset +2*ETHERADDRL),
			  (struct ether_addr *)(buffer->body + offset),
			  buffer->body + offset + sizeof(struct ether_header), 
			  maxlength);

}

long
fi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor_address(
  struct Hfi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor *adaptor_handle,
  HArrayOfByte *addr_handle)
{
  Classfi_hut_tcm_jacob_tcpip_ethernet_EthernetAdaptor *adaptor = 
    unhand(adaptor_handle);
  int fd = adaptor->handle;
  ArrayOfByte *addr = unhand(addr_handle);

  if (obj_length(addr_handle) != ETHERADDRL)
    SignalError(0, "java/lang/IllegalArgumentException",
	"Ethernet addr return table length must be exactly ETHER_ADDR_LEN");

  return ethernet_address(fd, (struct ether_addr *)addr->body);
}

/*
 * Throw an exception
 */

static void
throw_exception(char *string)
{
  static char buffer[BUFSIZ];

  sprintf(buffer, "%s: %s", string, strerror(errno));
  SignalError(0, "Jacob/Framework/EthernetException", buffer);
}

/**
 * Open opens the device, attaches the ppa and binds the sap.
 */
static int
ethernet_open(char *device, int ppa, int sap, int promisc)
{
  int retval = 0;
  int ether_fd;
  union DL_primitives rcvbuf;

  printf("Ethernet open: %s, %d, %x, %d\n", device, ppa, sap, promisc);

  if ((ether_fd = open(device, O_RDWR)) < 0) {
    throw_exception("ethernet_open: open system call");
    return (-1); 
  }

  if ((retval = dlpi_attach_req(ether_fd, ppa)) < 0) {
    ethernet_close_internal(ether_fd, DL_UNATTACHED);
    throw_exception("ethernet_open: dlpi attach");
    return retval;
  }

  if ((retval = dlpi_recv_ack(ether_fd, DL_OK_ACK, &rcvbuf, NULL, NULL)) < 0) {
    ethernet_close_internal(ether_fd, DL_UNATTACHED);
    throw_exception("ethernet_open: dlpi attach");
    return retval;
  }
      
  if ((retval = dlpi_bind_req(ether_fd, sap)) < 0) {
    ethernet_close_internal(ether_fd, DL_UNBOUND);
    throw_exception("ethernet_open: dlpi bind");
    return retval;
  }

  if ((retval = dlpi_recv_ack(ether_fd, DL_BIND_ACK, &rcvbuf, NULL, NULL))<0) {
    ethernet_close_internal(ether_fd, DL_UNBOUND);
    throw_exception("ethernet_open: dlpi bind");
    return retval;
  }

  if (promisc) {
    if ((retval = dlpi_promiscon_req(ether_fd)) < 0) {
      ethernet_close_internal(ether_fd, DL_IDLE);
      throw_exception("ethernet_open: dlpi promisc on");
      return retval;
    }

    if ((retval = dlpi_recv_ack(ether_fd, DL_OK_ACK, &rcvbuf, NULL, NULL))<0) {
       ethernet_close_internal(ether_fd, DL_IDLE);
       throw_exception("ethernet_open: dlpi promisc on");
       return retval;
    }
  }

  return ether_fd;
}

/**
 * 
 */

static void
ethernet_close(int ether_fd)
{
  if (ethernet_close_internal(ether_fd, DL_IDLE) < 0)
    throw_exception("ethernet_close");
}

/**
 * Closes the connection to the ethernet driver. The state
 * paramter is used to determine which actions have to be taken
 * when closing the connection (unbinding, detaching and closing).
 */

static int 
ethernet_close_internal(int ether_fd, int state)
{
  union DL_primitives rcvbuf;

  switch (state) {
  case DL_IDLE:
    dlpi_unbind_req(ether_fd);
    dlpi_recv_ack(ether_fd, DL_OK_ACK, &rcvbuf, NULL, NULL);
    /*FALLTHROUGH*/
  case DL_UNBOUND:
    dlpi_detach_req(ether_fd);
    dlpi_recv_ack(ether_fd, DL_OK_ACK, &rcvbuf, NULL, NULL);
    /*FALLTHROUGH*/
  case DL_UNATTACHED:
    return close(ether_fd);
  }
  return -1;
}

/**
 * Attaches the ppa to a open stream.
 */

static int 
dlpi_attach_req(int fd, int ppa)
{
  struct strbuf ctlbuf;
  dl_attach_req_t attach_req;

  attach_req.dl_primitive = DL_ATTACH_REQ;
  attach_req.dl_ppa = ppa;

  ctlbuf.len = sizeof (dl_attach_req_t);
  ctlbuf.buf = (char *) &attach_req;
  
  return putmsg(fd, &ctlbuf, NULL, 0);
}


/**
 * Detaches the stream from the current ppa
 */

static int 
dlpi_detach_req(int fd)
{
  struct strbuf ctlbuf;
  dl_detach_req_t detach_req;

  detach_req.dl_primitive = DL_DETACH_REQ;

  /* ctlbuf.len = sizeof (union DL_primitives); XXX ??? */
  ctlbuf.len = sizeof (dl_detach_req_t);
  ctlbuf.buf = (char *) &detach_req;
  
  return putmsg(fd, &ctlbuf, NULL, 0);
}


/**
 * Binds a sap.
 */

static int 
dlpi_bind_req(int fd, int sap)
{
  struct strbuf ctlbuf;
  dl_bind_req_t bind_req;

  bind_req.dl_primitive = DL_BIND_REQ;
  bind_req.dl_sap = sap;
  bind_req.dl_max_conind = 0;
  bind_req.dl_service_mode = DL_CLDLS;
  bind_req.dl_conn_mgmt = 0;
  bind_req.dl_xidtest_flg = 0;

  ctlbuf.len = sizeof (dl_bind_req_t);
  ctlbuf.buf = (char *) &bind_req;
  
  return putmsg(fd, &ctlbuf, NULL, 0);
}

/**
 * Unbinds the current sap.
 */

static int 
dlpi_unbind_req(int fd)
{
  dl_unbind_req_t unbind_req;  
  struct strbuf ctlbuf;

  unbind_req.dl_primitive = DL_UNBIND_REQ;

  ctlbuf.len = sizeof (dl_unbind_req_t);
  ctlbuf.buf = (char *) &unbind_req;
  
  return putmsg(fd, &ctlbuf, NULL, 0);
}

/*
 * Requests for promisc mode
 */

static int 
dlpi_promiscon_req(int fd)
{
  dl_promiscon_req_t promiscon_req;
  struct strbuf ctlbuf;

  promiscon_req.dl_primitive = DL_PROMISCON_REQ;
  promiscon_req.dl_level = DL_PROMISC_SAP;

  ctlbuf.len = sizeof (dl_promiscon_req_t);
  ctlbuf.buf = (char *) &promiscon_req;
  
  return putmsg(fd, &ctlbuf, NULL, 0);
}

/*
 * Issues the DLPI_INFO_REQ command
 *
 */

static int 
dlpi_info_req(int fd)
{
  struct strbuf ctlbuf;
  dl_info_req_t info_req;

  info_req.dl_primitive = DL_INFO_REQ;

  ctlbuf.len = sizeof (dl_info_req_t);
  ctlbuf.buf = (char *) &info_req;
  
  return putmsg(fd, &ctlbuf, NULL, RS_HIPRI);
}

/**
 * Requests for the ethernet address associated with the open stream
 */
 
static int 
dlpi_phys_addr_req(int fd)
{
  struct strbuf ctlbuf;
  dl_phys_addr_req_t addr_req;

  addr_req.dl_primitive = DL_PHYS_ADDR_REQ;
  addr_req.dl_addr_type = DL_CURR_PHYS_ADDR;

  ctlbuf.len = sizeof (dl_phys_addr_req_t);
  ctlbuf.buf = (char *) &addr_req;
  
  return putmsg(fd, &ctlbuf, NULL, RS_HIPRI);
}


/**
 * Receives answers on dlpi requests.
 */

static int 
dlpi_recv_ack(int fd, int okprim, union DL_primitives *rcvctl, 
	      char *rcvdata, int *rcvdatalen)
{
  struct strbuf ctlbuf;
  struct strbuf databuf;
  int retval = 0;
  int flags = 0;

  ctlbuf.maxlen = sizeof (*rcvctl);
  ctlbuf.len = 0;
  ctlbuf.buf = (char *) rcvctl;

  databuf.maxlen = 0;
  databuf.len    = 0;
  databuf.buf    = NULL;
  if (rcvdata != NULL && rcvdatalen != NULL) {
    databuf.maxlen = *rcvdatalen;
    databuf.buf    = rcvdata;
  }

  if ((retval = getmsg (fd, &ctlbuf, &databuf, &flags)) < 0)
    return retval;

  if (rcvdatalen != NULL) 
    *rcvdatalen = databuf.len;

  if (rcvctl->dl_primitive == okprim) 
    return retval;

  return -1;
}


/*
 * Prints information about the ethernet driver.
 */

static void
print_le_info(int ether_fd)
{
  union DL_primitives rcvbuf;
  dl_info_ack_t *info_ack = &(rcvbuf.info_ack);

  dlpi_info_req(ether_fd);
  dlpi_recv_ack(ether_fd, DL_INFO_ACK, &rcvbuf, NULL, NULL);

  printf("DL_INFO_ACK:\n");
  printf("dl_primitive %ld\n", info_ack->dl_primitive);
  printf("dl_max_sdu %ld\n", info_ack->dl_max_sdu);
  printf("dl_min_sdu %ld\n", info_ack->dl_min_sdu);
  printf("dl_addr_length %ld\n", info_ack->dl_addr_length);
  printf("dl_mac_type %ld\n", info_ack->dl_mac_type);
  printf("dl_reserved %ld\n", info_ack->dl_reserved);
  printf("dl_current_state %ld\n", info_ack->dl_current_state);
  printf("dl_sap_length %ld\n", info_ack->dl_sap_length);
  printf("dl_service_mode %ld\n", info_ack->dl_service_mode);
  printf("dl_qos_length %ld\n", info_ack->dl_qos_length);
  printf("dl_qos_offset %ld\n", info_ack->dl_qos_offset);
  printf("dl_qos_range_length %ld\n", info_ack->dl_qos_range_length);
  printf("dl_qos_range_offset %ld\n", info_ack->dl_qos_range_offset);
  printf("dl_provider_style %ld\n", info_ack->dl_provider_style);
  printf("dl_addr_offset %ld\n", info_ack->dl_addr_offset);
  printf("dl_version %ld\n", info_ack->dl_version);
  printf("dl_brdcst_addr_length %ld\n", info_ack->dl_brdcst_addr_length);      
  printf("dl_brdcst_addr_offset %ld\n", info_ack->dl_brdcst_addr_offset);
  printf("dl_growth %ld\n", info_ack->dl_growth);

  return;
}

/*
 * Prints out the ethernet address
 *
 */

static void 
print_phys_addr(int ether_fd)
{
  union DL_primitives rcvbuf;
  dl_phys_addr_ack_t *phys_addr = &(rcvbuf.physaddr_ack);
  unsigned char *ptr = (unsigned char *) phys_addr;
  int i;

  dlpi_phys_addr_req(ether_fd);
  dlpi_recv_ack(ether_fd, DL_PHYS_ADDR_ACK, &rcvbuf, NULL, NULL);

  printf("PHYS_ADDR:\n");
  printf("dl_primitive %ld\n", phys_addr->dl_primitive);
  printf("dl_addr_length %ld\n", phys_addr->dl_addr_length);
  printf("dl_addr_offset %ld\n", phys_addr->dl_addr_offset);
  ptr += sizeof(dl_phys_addr_ack_t);
  for(i = 0; i < sizeof(struct ether_addr); i++)
    printf("%x:",(unsigned char) *ptr++);
  printf("\n");
}


/*
 * Sends one packet to the ethernet driver with
 */

static int 
ethernet_send(int ether_fd, int ethertype, struct ether_addr *addr,
	      unsigned char *data, int dlen)
{
  struct strbuf ctlbuf;
  struct strbuf databuf;
  udata_packet_t udata;

  /// YYY
  udata.unitdata_req.dl_primitive = DL_UNITDATA_REQ;
  udata.unitdata_req.dl_dest_addr_length = sizeof(struct ether_addr) 
                                             + sizeof(u_short);
  udata.unitdata_req.dl_dest_addr_offset = sizeof(dl_unitdata_req_t);
  udata.unitdata_req.dl_priority.dl_min = 0;
  udata.unitdata_req.dl_priority.dl_max = 0;

  memcpy(udata.eth_addr.ether_addr_octet, addr, sizeof(struct ether_addr));
  udata.ether_type = ethertype;

  ctlbuf.len = sizeof(udata_packet_t);
  ctlbuf.buf = (char *) &udata;

  databuf.len = dlen;
  databuf.buf = (char *) data;

#if 0
  print_packet((unsigned char *) data, dlen, (unsigned char *) &udata, 
	       ctlbuf.len);
#endif

  return putmsg(ether_fd, &ctlbuf, &databuf, 0);
}


static void 
print_packet(unsigned char *ptr, int len, unsigned char *ptr2, int clen)
{
  int i;

  for(i = 0;i<clen;i++)
    printf( "%x ", *ptr2++);

  printf("\n");
  for(i = 0;i<len;i++)
    printf( "%x ", *ptr++);
  printf("\n");
  return;
}

/*
 * Gets one DL_UNITDATA_IND from the ethernet driver
 */

static int 
ethernet_receive(int ether_fd, 
		 unsigned short *ethertype,
		 struct ether_addr *addrs,
		 unsigned char *buffer, int maxlength)
{
  struct strbuf ctlbuf;
  struct strbuf databuf;
  union DL_primitives rcvbuf;
  dl_unitdata_ind_t *udata = &(rcvbuf.unitdata_ind);
  int flags = 0;
  int retval = 0;
  int x = TRUE;

  ctlbuf.maxlen = sizeof (union DL_primitives);
  ctlbuf.len = 0;
  ctlbuf.buf = (char *) &rcvbuf;

  databuf.maxlen = maxlength;
  databuf.len = 0;
  databuf.buf = (char *) buffer;
	  
  if ((retval = getmsg (ether_fd, &ctlbuf, &databuf, &flags)) < 0)
    return (retval);

  /* XXX */
  memcpy((unsigned char *)&addrs[0],
	 (unsigned char *)&rcvbuf + udata->dl_dest_addr_offset,
	 ETHERADDRL > udata->dl_dest_addr_length? 
	       udata->dl_dest_addr_length: ETHERADDRL);
  memcpy((unsigned char *)&addrs[1], 
	 (unsigned char *)&rcvbuf + udata->dl_src_addr_offset,
	 ETHERADDRL > udata->dl_src_addr_length? 
	       udata->dl_src_addr_length: ETHERADDRL);
  /* XXX */
  /*
  *ethertype = *((unsigned short *)(((unsigned char *)&rcvbuf)
				    + udata->dl_src_addr_offset 
				    + udata->dl_src_addr_length));
  */
  *ethertype = *((unsigned short *)(((unsigned char *)&udata)
				    + DL_ETHER_TYPE_OFFSET));


#if 0
  printf("Ethertype seems to be: %hd decimal, %hx hexadecimal\n", *ethertype, *ethertype);
  print_packet((unsigned char *) databuf.buf, databuf.len, (unsigned char *) &udata, 
	       ctlbuf.len);
#endif

  return databuf.len;
}

/*
 * Gets local ethernet address
 */

static int
ethernet_address(int ether_fd, struct ether_addr *addr)
{
  union DL_primitives rcvbuf;
  dl_phys_addr_ack_t *phys_addr = &(rcvbuf.physaddr_ack);

  dlpi_phys_addr_req(ether_fd);
  dlpi_recv_ack(ether_fd, DL_PHYS_ADDR_ACK, &rcvbuf, NULL, NULL);

  if (phys_addr->dl_addr_length != ETHERADDRL)
    return -1;

  memcpy((unsigned char *)addr,
	 (unsigned char *)phys_addr + phys_addr->dl_addr_offset,
	 phys_addr->dl_addr_length);
  return ETHERADDRL;
}

